import functools
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from typing import Generator, List, Union

import pandas as pd

from kabutobashi.domain.errors import KabutobashiPageError
from kabutobashi.domain.services import StockInfoHtmlDecoder
from kabutobashi.domain.values import IStockRecordsetRepository, StockInfoHtmlPage, StockRecordset

logger = getLogger(__name__)


__all__ = ["StockRecordsetStorageBasicRepository", "StockRecordsetCrawler"]


class StockRecordsetStorageBasicRepository(IStockRecordsetRepository):
    def __init__(self, path_candidate: Union[str, list], use_mp: bool = False, max_workers: int = 2):
        self.path_candidate = path_candidate
        self.use_mp = use_mp
        self.max_workers = max_workers

    def _read_path_generator(self) -> Generator[str, None, None]:
        if type(self.path_candidate) is str:
            yield self.path_candidate
        elif type(self.path_candidate) is list:
            for path in self.path_candidate:
                yield path
        else:
            raise ValueError()

    def _stock_recordset_read(self) -> StockRecordset:
        df_list = []
        if self.use_mp:
            with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                map_gen = executor.map(pd.read_csv, self._read_path_generator())
                for response in map_gen:
                    df_list.append(response)
        else:
            df_list = [pd.read_csv(p) for p in self._read_path_generator()]

        df = pd.concat(df_list)
        return StockRecordset.of(df=df)

    def _write_path_generator(self) -> Generator[str, None, None]:
        if type(self.path_candidate) is str:
            yield self.path_candidate
        elif type(self.path_candidate) is list:
            for path in self.path_candidate:
                yield path
        else:
            raise ValueError()

    def _stock_recordset_write(self, data: StockRecordset):
        df = data.to_df(minimum=False)
        for p in self._write_path_generator():
            df.to_csv(p, index=False)


class StockRecordsetCrawler(IStockRecordsetRepository):
    def __init__(self, code_list: list, dt: str, use_mp: bool = False, max_workers: int = 2):
        self.code_list = code_list
        self.dt = dt
        self.use_mp = use_mp
        self.max_workers = max_workers

    def _stock_recordset_read(self) -> StockRecordset:

        # 日次の株データ取得
        stock_data: List[dict] = []
        if self.use_mp:
            stock_data.extend(self.crawl_multiple(code_list=self.code_list, max_workers=self.max_workers, dt=self.dt))
        else:
            stock_data.extend([self.crawl_single(code=code, dt=self.dt) for code in self.code_list])

        # データを整形してStockDataとして保存
        df = pd.DataFrame(stock_data)
        return StockRecordset.of(df=df)

    @staticmethod
    def crawl_single(code: Union[int, str], dt: str) -> dict:
        try:
            stock_page_html = StockInfoHtmlPage.of(code=code, dt=dt)
            result = StockInfoHtmlDecoder(html_page=stock_page_html).decode()
            return result
        except KabutobashiPageError:
            return {}
        except AttributeError:
            logger.exception(f"error occurred at: {code}")
            return {}
        except Exception:
            logger.exception(f"error occurred at: {code}")
            return {}

    @staticmethod
    def crawl_multiple(code_list: List[Union[int, str]], dt: str, max_workers: int = 2) -> List[dict]:
        partial_crawl_single = functools.partial(StockRecordsetCrawler.crawl_single, dt=dt)
        response_list = []
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            map_gen = executor.map(partial_crawl_single, code_list)
            for response in map_gen:
                response_list.append(response)
        return response_list

    def _stock_recordset_write(self, data: StockRecordset):
        raise NotImplementedError()



from dataclasses import dataclass

import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


@dataclass(frozen=True)
class VolatilityProcess(ProcessMethod):
    """
    変動幅を計算する
    """

    method_name: str = "volatility"
    method_type: MethodType = MethodType.PARAMETERIZE

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["diff"] = -1
        # 正負が交差した点
        df_ = df_.join(self._cross(df_["diff"]))
        df_ = df_.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return df_

    def _processed_columns(self) -> list:
        return []

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        df = df_x.copy()
        df["volatility"] = (df["high"] - df["low"]) / df["close"]
        volatility_ = df["volatility"].mean()
        close_volatility = max(df["close"]) - min(df["close"]) / df["close"].median()
        return {"volatility": volatility_, "close_volatility": close_volatility}


@dataclass(frozen=True)
class VolatilityVisualize(VisualizeMethod):
    def _color_mapping(self) -> list:
        return []

    def _visualize_option(self) -> dict:
        return {"position": "-"}


volatility = Method.of(process_method=VolatilityProcess(), visualize_method=VolatilityVisualize())



from dataclasses import dataclass
from functools import reduce
from logging import getLogger
from typing import List, Optional, Union

import pandas as pd
import requests  # type: ignore
from bs4 import BeautifulSoup

from kabutobashi.domain.entity import StockIpo, Weeks52HighLow
from kabutobashi.domain.values import (
    StockInfoHtmlPage,
    StockInfoMultipleDaysMainHtmlPage,
    StockInfoMultipleDaysSubHtmlPage,
    StockIpoHtmlPage,
    StockWeeks52HighLowHtmlPage,
)

logger = getLogger(__name__)


@dataclass(frozen=True)
class PageDecoder:
    tag1: Optional[str] = None
    class1: Optional[str] = None
    id1: Optional[str] = None
    default: str = ""

    def _decode(self, value):
        class1 = {"class": self.class1}

        set_value = None
        # tag1から取得
        if self.tag1 is not None:
            if class1["class"] is not None:
                set_value = value.find(self.tag1, self.class1)
            else:
                set_value = value.find(self.tag1)

        if set_value is None:
            return self.default

        # 文字列を置換して保持
        return self.replace(set_value.get_text())

    def decode(self, bs: BeautifulSoup) -> Union[str, List[str]]:
        return self._decode(value=bs)

    @staticmethod
    def replace(input_text: str) -> str:
        target_list = [" ", "\t", "\n", "\r", "円"]

        def remove_of(_input: str, target: str):
            return _input.replace(target, "")

        result = reduce(remove_of, target_list, input_text)
        return result.replace("\xa0", " ")


@dataclass(frozen=True)
class StockInfoHtmlDecoder:
    """

    Examples:
        >>> from kabutobashi import StockInfoHtmlPage
        >>> # get single page
        >>> page_html = StockInfoHtmlPage.of(code="0001", dt="2022-07-22")
        >>> result = StockInfoHtmlDecoder(page_html=page_html).decode()
    """

    html_page: StockInfoHtmlPage

    def decode(self) -> dict:
        soup = self.html_page.get_as_soup()
        result = {}

        stock_board_tag = "md_stockBoard"

        # ページ上部の情報を取得
        stock_board = soup.find("div", {"class": stock_board_tag})
        result.update(
            {
                "stock_label": PageDecoder(tag1="div", class1="stock_label").decode(bs=stock_board),
                "name": PageDecoder(tag1="p", class1="md_stockBoard_stockName").decode(bs=stock_board),
                "close": PageDecoder(tag1="div", class1="stock_price").decode(bs=stock_board),
                # "date": PageDecoder(tag1="h2", class1="stock_label fsl").decode(bs=stock_board),
            }
        )

        # ページ中央の情報を取得
        stock_detail = soup.find("div", {"id": "main"})
        info = {}
        for li in stock_detail.find_all("tr", {"class": "ly_vamd"}):
            info[li.find("th").get_text()] = li.find("td").get_text()
        result.update(
            {
                "dt": self.html_page.dt,
                "code": str(self.html_page.code),
                "industry_type": PageDecoder(tag1="div", class1="ly_content_wrapper size_ss").decode(bs=stock_detail),
                "open": info.get("始値", "0"),
                "high": info.get("高値", "0"),
                "low": info.get("安値", "0"),
                "unit": info.get("単元株数", "0"),
                "per": info.get("PER(調整後)", "0"),
                "psr": info.get("PSR", "0"),
                "pbr": info.get("PBR", "0"),
                "volume": info.get("出来高", "0"),
                "market_capitalization": info.get("時価総額", "---"),
                "issued_shares": info.get("発行済株数", "---"),
                "is_delisting": False,
            }
        )

        # 上場廃止の確認
        open_value = result["open"]
        high_value = result["high"]
        low_value = result["low"]
        if open_value == "---" and high_value == "---" and low_value == "---":
            result.update({"is_delisting": True})
        return result


@dataclass(frozen=True)
class StockIpoHtmlDecoder:

    html_page: StockIpoHtmlPage

    def decode(self) -> dict:
        soup = self.html_page.get_as_soup()
        table_content = soup.find("div", {"class": "tablewrap"})
        table_thead = table_content.find("thead")
        # headの取得
        table_head_list = []
        for th in table_thead.find_all("th"):
            table_head_list.append(th.get_text())

        # bodyの取得
        table_tbody = table_content.find("tbody")
        whole_result = []
        for idx, tr in enumerate(table_tbody.find_all("tr")):
            table_body_dict = {}
            for header, td in zip(table_head_list, tr.find_all("td")):
                table_body_dict[header] = td.get_text().replace("\n", "")
            whole_result.append(StockIpo.loads(data=table_body_dict).dumps())
        return {"ipo_list": whole_result}


@dataclass(frozen=True)
class Weeks52HighLowHtmlDecoder:
    html_page: StockWeeks52HighLowHtmlPage

    def decode(self) -> dict:
        soup = self.html_page.get_as_soup()

        content = soup.find("body").find("tbody")
        table = content.find_all("tr")
        whole_result = []
        for t in table:
            buy_or_sell = ""

            for td in t.find_all("td"):
                if td.text in ["買い", "強い買い", "売り", "強い売り"]:
                    buy_or_sell = td.text
            data = {
                "code": PageDecoder(tag1="a").decode(bs=t),
                "brand_name": PageDecoder(tag1="span").decode(bs=t),
                "buy_or_sell": buy_or_sell,
                "dt": self.html_page.dt,
            }
            whole_result.append(Weeks52HighLow.from_page_of(data=data).dumps())

        return {"weeks_52_high_low": whole_result}


@dataclass(frozen=True)
class StockInfoMultipleDaysHtmlDecoder:
    """

    Examples:
        >>> import kabutobashi as kb
        >>> main_html_page = kb.StockInfoMultipleDaysMainHtmlPage.of(1375)
        >>> sub_html_page = kb.StockInfoMultipleDaysSubHtmlPage.of(1375)
        >>> data = kb.StockInfoMultipleDaysHtmlDecoder(main_html_page, sub_html_page).decode()
        >>> df = pd.DataFrame(data)
        >>> records = kb.StockRecordset.of(df)
    """

    main_html_page: StockInfoMultipleDaysMainHtmlPage
    sub_html_page: StockInfoMultipleDaysSubHtmlPage

    def decode(self) -> dict:
        result_1 = []
        result_2 = []
        main_soup = self.main_html_page.get_as_soup()
        sub_soup = self.sub_html_page.get_as_soup()
        stock_recordset_tag = "md_card md_box"

        # ページの情報を取得
        stock_recordset = main_soup.find("div", {"class": stock_recordset_tag})
        # code,market,name,industry_type,open,high,low,close,psr,per,pbr,volume,unit,market_capitalization,issued_shares,dt,crawl_datetime
        mapping = {0: "dt", 1: "open", 2: "high", 3: "low", 4: "close", 5: "調整後終値", 6: "volume"}
        for tr in stock_recordset.find_all("tr"):
            tmp = {}
            for idx, td in enumerate(tr.find_all("td")):
                tmp.update({mapping[idx]: td.get_text()})
            result_1.append(tmp)

        # そのほかの情報
        # soup = BeautifulSoup(get_url_text(url=valuation_url.format(code=code)), features="lxml")
        stock_board = sub_soup.find("div", {"class": "md_card md_box mzp"})
        mapping2 = {0: "dt", 1: "psr", 2: "per", 3: "pbr", 4: "配当利回り(%)", 5: "close", 6: "調整後終値", 7: "volume"}
        for tr in stock_board.find_all("tr"):
            tmp = {}
            for idx, td in enumerate(tr.find_all("td")):
                tmp.update({mapping2[idx]: td.get_text()})
            result_2.append(tmp)

        df1 = pd.DataFrame(result_1).dropna()
        df2 = pd.DataFrame(result_2).dropna()

        df1 = df1[["dt", "open", "high", "low", "close"]]
        df2 = df2[["dt", "psr", "per", "pbr", "volume"]]

        df = pd.merge(df1, df2, on="dt")
        df["code"] = self.main_html_page.code
        return df.to_dict(orient="records")



from dataclasses import dataclass

import numpy as np
import pandas as pd
from scipy.optimize import curve_fit

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


@dataclass(frozen=True)
class FittingProcess(ProcessMethod):
    """
    1次、2次、3次の関数でfittingした値を返す
    """

    method_name: str = "fitting"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    @staticmethod
    def _linear_fit(x, a, b):
        return a * x + b

    @staticmethod
    def _square_fit(x, a, b, c):
        return a * x * x + b * x + c

    @staticmethod
    def _cube_fit(x, a, b, c, d):
        return a * x * x * x + b * x * x + c * x + d

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        array_y = df["close"]
        array_x = np.array(range(0, len(array_y)))

        linear_param, _ = curve_fit(self._linear_fit, array_x, array_y)
        square_param, _ = curve_fit(self._square_fit, array_x, array_y)
        cube_param, _ = curve_fit(self._cube_fit, array_x, array_y)
        df_ = df.copy()
        df_["linear_fitting"] = [self._linear_fit(x, *linear_param) for x in array_x]
        df_["square_fitting"] = [self._square_fit(x, *square_param) for x in array_x]
        df_["cube_fitting"] = [self._cube_fit(x, *cube_param) for x in array_x]
        return df_

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["diff"] = -1
        # 正負が交差した点
        df_ = df_.join(self._cross(df_["diff"]))
        df_ = df_.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return df_

    def _processed_columns(self) -> list:
        return ["linear_fitting", "square_fitting", "cube_fitting"]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        array_y = df_x["close"]
        array_x = np.array(range(0, len(array_y)))

        linear_param, _ = curve_fit(self._linear_fit, array_x, array_y)
        square_param, _ = curve_fit(self._square_fit, array_x, array_y)
        cube_param, _ = curve_fit(self._cube_fit, array_x, array_y)

        return {
            "linear_x_1": linear_param[0],
            "linear_x_0": linear_param[1],
            "square_x_2": square_param[0],
            "square_x_1": square_param[1],
            "square_x_0": square_param[2],
            "cube_x_3": cube_param[0],
            "cube_x_2": cube_param[1],
            "cube_x_1": cube_param[2],
            "cube_x_0": cube_param[3],
        }


@dataclass(frozen=True)
class FittingVisualize(VisualizeMethod):
    """
    1次、2次、3次の関数でfittingした値を返す
    """

    def _color_mapping(self) -> list:
        return [
            {"df_key": "linear_fitting", "color": "#dc143c", "label": "linear"},
            {"df_key": "square_fitting", "color": "#ffa500", "label": "square"},
            {"df_key": "cube_fitting", "color": "#1e90ff", "label": "cube"},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "in"}


fitting = Method.of(process_method=FittingProcess(), visualize_method=FittingVisualize())



from dataclasses import dataclass, field
from typing import Any, Dict, List

import matplotlib.pyplot as plt
import pandas as pd

from kabutobashi.domain.errors import KabutobashiEntityError


@dataclass(frozen=True)
class StockDataProcessed:
    """
    StockDataProcessedBySingleMethod: ValueObject
    Holds data processed by singular-Method.
    """

    code: str
    start_at: str
    end_at: str
    applied_method_name: str
    df: pd.DataFrame = field(repr=False)
    df_required_columns: List[str] = field(repr=False)
    parameters: Dict[str, Any]

    def __post_init__(self):
        df_columns = self.df.columns
        if not all([c in df_columns for c in self.df_required_columns]):
            raise KabutobashiEntityError()

    def get_impact(self, influence: int = 2, tail: int = 5) -> Dict[str, float]:
        """

        Args:
            influence:
            tail:

        Returns:
            Dict[str, float]

        Examples:
        """
        return {self.applied_method_name: self._get_impact(df=self.df, influence=influence, tail=tail)}

    @staticmethod
    def _get_impact(df: pd.DataFrame, influence: int, tail: int) -> float:
        """
        売りと買いのシグナルの余波の合計値を返す。

        Args:
            df:
            influence:
            tail:

        Returns:
            [-1,1]の値をとる。-1: 売り、1: 買いを表す
        """
        df["buy_impact"] = df["buy_signal"].ewm(span=influence).mean()
        df["sell_impact"] = df["sell_signal"].ewm(span=influence).mean()
        buy_impact_index = df["buy_impact"].iloc[-tail:].sum()
        sell_impact_index = df["sell_impact"].iloc[-tail:].sum()
        return round(buy_impact_index - sell_impact_index, 5)


@dataclass(frozen=True)
class StockDataVisualized:
    """
    StockDataVisualized: ValueObject
    Used to visualize.
    """

    fig: plt.Figure
    size_ratio: int

    def __post_init__(self):
        if type(self.fig) is not plt.Figure:
            raise KabutobashiEntityError()


@dataclass(frozen=True)
class StockDataEstimated:
    """
    StockDataEstimatedBySingleFilter: ValueObject
    """

    code: str
    estimated_value: float
    estimate_filter_name: str

    def weighted_estimated_value(self, weights: dict) -> float:
        weight = weights.get(self.estimate_filter_name, 1)
        return weight * self.estimated_value



import pytest

import kabutobashi as kb


def test_crawl_page_detail():
    result = kb.StockRecordsetCrawler.crawl_single(code=4395, dt="2022-07-23")
    assert result is not None
    assert type(result) is dict

    result = kb.StockRecordsetCrawler(code_list=[4395], dt="2022-07-23").read()
    assert result is not None
    assert type(result) is kb.StockRecordset


def test_crawl_ipo_list():
    html_page = kb.StockIpoHtmlPage.of(year="2019")
    result = kb.StockIpoHtmlDecoder(html_page=html_page).decode()
    assert result is not None
    assert type(result) is dict


def test_crawl_week_52_high_low_list():
    html_page = kb.StockWeeks52HighLowHtmlPage.of(data_type="newly_low", dt="2022-07-23")
    result = kb.Weeks52HighLowHtmlDecoder(html_page=html_page).decode()
    assert result is not None
    assert type(result) is dict



from dataclasses import dataclass

import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


@dataclass(frozen=True)
class AdxProcess(ProcessMethod):
    """
    相場のトレンドの強さを見るための指標である`ADX`を計算するクラス。

    以下の指標を計算するクラス

    * +DI: 株価の上昇の大きさ
    * -DI: 株価の下降の大きさ
    * ADX: 株価のトレンドの強さ
    * ADXR: ADXの単純移動平均線

    Args:
        term (int):
        adx_term (int):
        adxr_term (int):

    See Also:
        * https://www.sevendata.co.jp/shihyou/technical/dmi.html
        * https://www.sevendata.co.jp/shihyou/technical/adx.html

    """

    term: int = 14
    adx_term: int = 14
    adxr_term: int = 28
    method_name: str = "adx"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    @staticmethod
    def _true_range(x: pd.DataFrame):
        """

        Args:
            x (pd.DataFrame)

        Returns:
            maximum
        """
        if x["high"] is pd.NA or x["low"] is pd.NA or x["shift_close"] is pd.NA:
            return 0
        current_high = x["high"]
        current_low = x["low"]
        prev_close = x["shift_close"]
        a = current_high - current_low
        b = current_high - prev_close
        c = prev_close - current_low
        max_ab = max(a, b)
        max_ac = max(a, c)
        return max(max_ab, max_ac)

    @staticmethod
    def _compute_dx(x: pd.DataFrame) -> float:
        """

        Args:
            x (pd.DataFrame):
        """
        numerator = abs(x["plus_di"] - x["minus_di"])
        denominator = x["plus_di"] + x["minus_di"]
        try:
            return numerator / denominator * 100
        except ZeroDivisionError:
            return 0

    @staticmethod
    def _fixed_plus_dm(x: pd.DataFrame) -> float:
        if x["plus_dm"] is pd.NA or x["minus_dm"] is pd.NA:
            return 0
        if x["plus_dm"] > 0 and x["plus_dm"] > x["minus_dm"]:
            return x["plus_dm"]
        else:
            return 0

    @staticmethod
    def _fixed_minus_dm(x: pd.DataFrame) -> float:
        if x["plus_dm"] is pd.NA or x["minus_dm"] is pd.NA:
            return 0
        if x["minus_dm"] > 0 and x["minus_dm"] > x["plus_dm"]:
            return x["minus_dm"]
        else:
            return 0

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        # 利用する値をshift
        df = df.assign(shift_high=df["high"].shift(1), shift_low=df["low"].shift(1), shift_close=df["close"].shift(1))
        df = df.assign(
            plus_dm=df.apply(lambda x: x["high"] - x["shift_high"], axis=1),
            minus_dm=df.apply(lambda x: x["shift_low"] - x["low"], axis=1),
        )
        df = df.assign(
            fixed_plus_dm=df.apply(lambda x: self._fixed_plus_dm(x), axis=1),
            fixed_minus_dm=df.apply(lambda x: self._fixed_minus_dm(x), axis=1),
        )
        df = df.assign(
            true_range=df.apply(lambda x: self._true_range(x), axis=1),
            sum_tr=lambda x: x["true_range"].rolling(self.term).sum(),
            sum_plus_dm=lambda x: x["fixed_plus_dm"].rolling(self.term).sum(),
            sum_minus_dm=lambda x: x["fixed_minus_dm"].rolling(self.term).sum(),
        )

        df = df.dropna()
        # +DI, -DI
        df = df.assign(
            plus_di=df.apply(lambda x: x["sum_plus_dm"] / x["sum_tr"] * 100, axis=1),
            minus_di=df.apply(lambda x: x["sum_minus_dm"] / x["sum_tr"] * 100, axis=1),
        )
        df = df.assign(
            DX=df.apply(self._compute_dx, axis=1),
            ADX=lambda x: x["DX"].rolling(self.adx_term).mean(),
            ADXR=lambda x: x["DX"].rolling(self.adxr_term).mean(),
        )
        return df

    @staticmethod
    def _buy_signal(x) -> float:
        """
        DMIとADXを組み合わせた基本パターン
        """

        # +DIが-DIを上抜き、ADXが上昇傾向の上向きであれば新規買い
        if x["ADX_trend"] > 0:
            if x["to_plus"] > 0:
                return 1
        return 0

        # +DIが-DIより上に位置している際に、
        # ADXが下向きから上向きに転換した場合
        # if x['diff'] > 0:

    @staticmethod
    def _sell_signal(x) -> float:
        """
        DMIとADXを組み合わせた基本パターン
        """
        # -DIが+DIを下抜き、ADXが下落傾向の下向きであれば新規空売り
        if x["ADX_trend"] < 0:
            if x["to_minus"] > 0:
                return 1
        return 0

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        buy_signalとsell_signalを付与
        """
        df["ADX_trend"] = self._trend(df["ADX"])
        df["diff"] = df["plus_di"] - df["minus_di"]
        df = df.join(self._cross(df["diff"]))

        df["buy_signal"] = df.apply(lambda x: self._buy_signal)
        df["sell_signal"] = df.apply(lambda x: self._sell_signal)

        return df

    def _processed_columns(self) -> list:
        return ["plus_di", "minus_di", "DX", "ADX", "ADXR"]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return {
            "adx_dx": df_p["DX"].tail(3).mean(),
            "adx_adx": df_p["ADX"].tail(3).mean(),
            "adx_adxr": df_p["ADXR"].tail(3).mean(),
        }


@dataclass(frozen=True)
class AdxVisualize(VisualizeMethod):
    """
    相場のトレンドの強さを見るための指標である`ADX`を計算するクラス。

    以下の指標を計算するクラス

    * +DI: 株価の上昇の大きさ
    * -DI: 株価の下降の大きさ
    * ADX: 株価のトレンドの強さ
    * ADXR: ADXの単純移動平均線

    Args:
        term (int):
        adx_term (int):
        adxr_term (int):

    See Also:
        * https://www.sevendata.co.jp/shihyou/technical/dmi.html
        * https://www.sevendata.co.jp/shihyou/technical/adx.html

    """

    def _color_mapping(self) -> list:
        return [
            {"df_key": "plus_di", "color": "", "label": "+DI"},
            {"df_key": "minus_di", "color": "", "label": "-DI"},
            {"df_key": "ADX", "color": "", "label": "ADX"},
            {"df_key": "ADXR", "color": "", "label": "ADXR"},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "lower"}


adx = Method.of(process_method=AdxProcess(), visualize_method=AdxVisualize())



"""
Method modules provide technical analysis for stock chart.

- technical analysis

  - ADX
  - BollingerBands
  - Fitting
  - Ichimoku
  - MACD
  - Momentum
  - PsychoLogical
  - SMA
  - Stochastics

- other

  - Basic: only used `parameterize`

"""
from .adx import AdxProcess, adx
from .basic import BasicProcess, basic
from .bollinger_bands import BollingerBandsProcess, bollinger_bands
from .fitting import FittingProcess, fitting
from .ichimoku import IchimokuProcess, ichimoku
from .industry_cat import IndustryCategoriesProcess, industry_categories
from .macd import MacdProcess, macd
from .method import Method, ProcessMethod
from .momentum import MomentumProcess, momentum
from .pct_change import PctChangeProcess, pct_change
from .psycho_logical import PsychoLogicalProcess, psycho_logical
from .sma import SmaProcess, sma
from .stochastics import StochasticsProcess, stochastics
from .volatility import VolatilityProcess, volatility



from .html_pages import (
    StockInfoHtmlPage,
    StockInfoMultipleDaysMainHtmlPage,
    StockInfoMultipleDaysSubHtmlPage,
    StockIpoHtmlPage,
    StockWeeks52HighLowHtmlPage,
)
from .stock_data import StockDataEstimated, StockDataProcessed, StockDataVisualized
from .stock_recordset import IStockRecordsetRepository, StockRecordset



from dataclasses import dataclass, field
from typing import Union

import requests  # type: ignore
from bs4 import BeautifulSoup

from kabutobashi.domain.errors import KabutobashiPageError

from .user_agent import UserAgent


@dataclass(frozen=True)
class HtmlPage:
    """
    id: code & dt
    """

    html: str = field(repr=False)
    page_type: str
    url: str

    def __post_init__(self):
        assert self.page_type in ["info", "info_multiple", "ipo", "weeks_52_high_low"]

    @staticmethod
    def from_url(url: str, page_type: str) -> "HtmlPage":
        """
        TODO repositoryを利用するように修正する
        requestsを使って、webからhtmlを取得する
        """
        user_agent = UserAgent.get_user_agent_header()
        r = requests.get(url, headers=user_agent)

        if r.status_code != 200:
            raise KabutobashiPageError(url=url)

        # 日本語に対応
        r.encoding = r.apparent_encoding
        return HtmlPage(html=r.text, page_type=page_type, url=url)

    def get_as_soup(self) -> BeautifulSoup:
        return BeautifulSoup(self.html, features="lxml")


@dataclass(frozen=True)
class StockInfoHtmlPage(HtmlPage):
    code: Union[int, str]
    dt: str

    @staticmethod
    def of(code: Union[int, str], dt: str) -> "StockInfoHtmlPage":
        url = f"https://minkabu.jp/stock/{code}"
        page_type = "info"
        html_page = HtmlPage.from_url(url=url, page_type=page_type)
        return StockInfoHtmlPage(page_type=page_type, code=code, dt=dt, html=html_page.html, url=url)


@dataclass(frozen=True)
class StockIpoHtmlPage(HtmlPage):
    year: str

    @staticmethod
    def of(year: str) -> "StockIpoHtmlPage":
        url = f"https://96ut.com/ipo/list.php?year={year}"
        page_type = "ipo"
        html_page = HtmlPage.from_url(url=url, page_type=page_type)
        return StockIpoHtmlPage(page_type=page_type, year=year, html=html_page.html, url=url)


@dataclass(frozen=True)
class StockWeeks52HighLowHtmlPage(HtmlPage):
    data_type: str
    dt: str

    def __post_init__(self):
        if self.data_type not in ["high", "low", "newly_high", "newly_low"]:
            raise KabutobashiPageError()

    @staticmethod
    def _url_suffix(data_type: str) -> str:
        # 52週の高値・底値を取得する関数とURL
        if data_type == "high":
            return "highs-and-lows-52wk-high"
        elif data_type == "low":
            return "highs-and-lows-52wk-low"
        elif data_type == "newly_high":
            return "highs-and-lows-ath"
        elif data_type == "newly_low":
            return "highs-and-lows-atl"

        raise KabutobashiPageError()

    @staticmethod
    def of(data_type: str, dt: str) -> "StockWeeks52HighLowHtmlPage":
        base_url = "https://jp.tradingview.com/markets/stocks-japan"
        url = f"{base_url}/{StockWeeks52HighLowHtmlPage._url_suffix(data_type=data_type)}"
        page_type = "weeks_52_high_low"
        html_page = HtmlPage.from_url(url=url, page_type=page_type)
        return StockWeeks52HighLowHtmlPage(
            page_type=page_type, data_type=data_type, html=html_page.html, url=url, dt=dt
        )


@dataclass(frozen=True)
class StockInfoMultipleDaysMainHtmlPage(HtmlPage):
    code: Union[int, str]

    @staticmethod
    def of(code: Union[int, str]) -> "StockInfoMultipleDaysMainHtmlPage":
        url = f"https://minkabu.jp/stock/{code}/daily_bar"
        page_type = "info_multiple"
        html_page = HtmlPage.from_url(url=url, page_type=page_type)
        return StockInfoMultipleDaysMainHtmlPage(page_type=page_type, code=code, html=html_page.html, url=url)


@dataclass(frozen=True)
class StockInfoMultipleDaysSubHtmlPage(HtmlPage):
    code: Union[int, str]

    @staticmethod
    def of(code: Union[int, str]) -> "StockInfoMultipleDaysSubHtmlPage":
        url = f"https://minkabu.jp/stock/{code}/daily_valuation"
        page_type = "info_multiple"
        html_page = HtmlPage.from_url(url=url, page_type=page_type)
        return StockInfoMultipleDaysSubHtmlPage(page_type=page_type, code=code, html=html_page.html, url=url)



from dataclasses import dataclass

import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


@dataclass(frozen=True)
class PsychoLogicalProcess(ProcessMethod):
    """
    See Also:
        https://www.sevendata.co.jp/shihyou/technical/psycho.html
    """

    upper_threshold: float = 0.75
    lower_threshold: float = 0.25
    psycho_term: float = 12
    method_name: str = "psycho_logical"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["shift_close"] = df_["close"].shift(1)
        df_ = df_.fillna(0)
        df_["diff"] = df_.apply(lambda x: x["close"] - x["shift_close"], axis=1)

        df_["is_raise"] = df_["diff"].apply(lambda x: 1 if x > 0 else 0)

        df_["psycho_sum"] = df_["is_raise"].rolling(self.psycho_term).sum()
        df_["psycho_line"] = df_["psycho_sum"].apply(lambda x: x / self.psycho_term)

        df_["bought_too_much"] = df_["psycho_line"].apply(lambda x: 1 if x > self.upper_threshold else 0)
        df_["sold_too_much"] = df_["psycho_line"].apply(lambda x: 1 if x < self.lower_threshold else 0)
        return df_

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df["buy_signal"] = df["sold_too_much"]
        df["sell_signal"] = df["bought_too_much"]
        return df

    def _processed_columns(self) -> list:
        return ["psycho_line", "bought_too_much", "sold_too_much"]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return {"psycho_line": df_p["psycho_line"].tail(3).mean()}


@dataclass(frozen=True)
class PsychoLogicalVisualize(VisualizeMethod):
    """
    See Also:
        https://www.sevendata.co.jp/shihyou/technical/psycho.html
    """

    def _color_mapping(self) -> list:
        return [
            {"df_key": "psycho_line", "color": "", "label": "psycho_line"},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "lower"}


psycho_logical = Method.of(process_method=PsychoLogicalProcess(), visualize_method=PsychoLogicalVisualize())



import kabutobashi as kb


def test_workday():
    date_list = kb.get_past_n_days("2020-03-31", 30)
    assert len(date_list) == 30
    # 祝日が除かれていることを確認
    assert "2020-03-20" not in date_list
    # 土日が除かれていることを確認
    assert "2020-03-07" not in date_list
    assert "2020-03-08" not in date_list
    assert "2020-03-14" not in date_list
    assert "2020-03-15" not in date_list
    assert "2020-03-21" not in date_list
    assert "2020-03-22" not in date_list
    assert "2020-03-28" not in date_list
    assert "2020-03-29" not in date_list



import pydantic
import pytest

import kabutobashi as kb
from kabutobashi.domain.errors import KabutobashiEntityError


@pytest.fixture(scope="module", autouse=True)
def example_records() -> kb.StockRecordset:
    yield kb.example()


class TestStockBrand:
    def test_error_init(self):
        with pytest.raises(pydantic.ValidationError):
            _ = kb.StockBrand(
                id=None,
                code="1234",
                market="",
                name="",
                unit="",
                market_capitalization="",
                industry_type="",
                issued_shares="",
            )


class TestStockRecord:
    def test_error_init(self):
        with pytest.raises(pydantic.ValidationError):
            _ = kb.StockRecord(
                id=None,
                code="1234",
                open="",
                high="",
                low="",
                close="",
                psr="",
                per="",
                pbr="",
                volume="",
                dt="",
                is_delisting=False,
            )


class TestStockIpo:
    def test_error_init(self):
        with pytest.raises(pydantic.ValidationError):
            _ = kb.StockIpo(
                id=None, code="", manager="", stock_listing_at="", public_offering="", evaluation="", initial_price=""
            )


class TestWeeks52HihLow:
    def test_error_init(self):
        with pytest.raises(pydantic.ValidationError):
            _ = kb.Weeks52HighLow(
                code="", brand_name="", close="", buy_or_sell="", volatility_ratio="", volatility_value=""
            )


class TestStockRecordset:
    def test_code_iterable(self, example_records: kb.StockRecordset):
        for _ in example_records.to_code_iterable(until=1):
            pass

    def test_multiple_code_error(self, example_records: kb.StockRecordset):
        with pytest.raises(KabutobashiEntityError):
            _ = example_records.get_single_code_recordset_status()

    def test_invalid_column_error(self, example_records: kb.StockRecordset):
        # check invalid column
        with pytest.raises(KabutobashiEntityError):
            _ = kb.StockRecordset.of(df=example_records.to_df()[["close"]])

    def test_get_df(self, example_records: kb.StockRecordset):
        records = example_records.to_single_code(code="1375")

        required_cols = ["code", "open", "close", "high", "low", "volume", "per", "psr", "pbr", "dt"]
        optional_cols = ["name", "industry_type", "market", "unit"]

        # check minimum df
        minimum_df = records.to_df()
        assert all([(c in minimum_df.columns) for c in required_cols])
        assert all([(c not in minimum_df.columns) for c in optional_cols])

        # check full df
        full_df = records.to_df(minimum=False)
        assert all([(c in full_df.columns) for c in required_cols])
        assert all([(c in full_df.columns) for c in optional_cols])

        latest_date_df = records.to_df(latest=True)
        assert len(latest_date_df.index) == 1

        status = records.get_single_code_recordset_status()
        assert status.code == "1375"


class TestStockSingleAggregate:
    def test_pass(self, example_records: kb.StockRecordset):
        methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility]
        agg = kb.StockCodeSingleAggregate.of(entity=example_records, code="1375")
        estimated = agg.with_processed(methods=methods).with_estimated(stock_analysis=kb.stock_analysis)
        value = estimated.weighted_estimated_value({"fundamental": 1.0, "volume": 1.0})
        assert value != 0

        # check visualize single column
        data_visualized = agg.visualize(kb.sma)
        assert data_visualized.fig

        # check visualize multiple columns
        data_visualized = agg.visualize(kb.macd)
        assert data_visualized.fig



# import seaborn as sns

# import errors
from kabutobashi.domain.services import SaFundamental, SaVolume, StockAnalysis

# methods to analysis
from kabutobashi.domain.services.method import (
    Method,
    adx,
    basic,
    bollinger_bands,
    fitting,
    ichimoku,
    industry_categories,
    macd,
    momentum,
    pct_change,
    psycho_logical,
    sma,
    stochastics,
    volatility,
)

from .domain import errors
from .domain.aggregates import StockCodeSingleAggregate
from .domain.entity import OPTIONAL_COL, REQUIRED_COL, StockBrand, StockIpo, StockRecord, Weeks52HighLow
from .domain.services import StockInfoMultipleDaysHtmlDecoder, StockIpoHtmlDecoder, Weeks52HighLowHtmlDecoder
from .domain.values import (
    StockDataEstimated,
    StockDataProcessed,
    StockDataVisualized,
    StockInfoHtmlPage,
    StockInfoMultipleDaysMainHtmlPage,
    StockInfoMultipleDaysSubHtmlPage,
    StockIpoHtmlPage,
    StockRecordset,
    StockWeeks52HighLowHtmlPage,
)
from .example_data import example
from .infrastructure.repository import StockRecordsetCrawler, StockRecordsetStorageBasicRepository

# n営業日前までの日付のリストを返す関数; 銘柄コードでイテレーションする関数; window幅でデータを取得しつつデータを返す関数; 株価の動きを様々な統計量で表現
from .utilities import get_past_n_days

# sns.set()

methods = [sma, macd, stochastics, adx, bollinger_bands, momentum, psycho_logical, fitting, basic]

# estimate filters
sa_fundamental = SaFundamental()
sa_volume = SaVolume()

stock_analysis = [sa_fundamental, sa_volume]

# comparable tuple
VERSION = (0, 5, 0)
# generate __version__ via VERSION tuple
__version__ = ".".join(map(str, VERSION))

# module level doc-string
__doc__ = """
kabutobashi
===========

**kabutobashi** is a Python package to analysis stock data with measure
analysis methods, such as MACD, SMA, etc.

Main Features
-------------
Here are the things that kabutobashi does well:
 - Easy crawl.
 - Easy analysis.
"""



from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass(frozen=True)  # type: ignore
class StockAnalysis(ABC):
    estimate_filter_name: str

    def estimate(self, data: dict) -> float:
        self._validate(data=data)
        return self._estimate(data=data)

    @abstractmethod
    def _validate(self, data: dict):
        raise NotImplementedError()

    @abstractmethod
    def _estimate(self, data: dict) -> float:
        raise NotImplementedError()


@dataclass(frozen=True)
class SaFundamental(StockAnalysis):
    estimate_filter_name: str = "fundamental"

    def _validate(self, data: dict):
        if "sma" not in data.keys():
            raise KeyError()
        if "macd" not in data.keys():
            raise KeyError()
        if "stochastics" not in data.keys():
            raise KeyError()
        if "bollinger_bands" not in data.keys():
            raise KeyError()
        if "momentum" not in data.keys():
            raise KeyError()
        if "psycho_logical" not in data.keys():
            raise KeyError()

    def _estimate(self, data: dict) -> float:
        return (
            data["sma"] * 1.5
            + data["macd"] * 1.1
            + data["stochastics"] * 0.2
            + data["bollinger_bands"]
            + data["momentum"]
            + data["psycho_logical"]
        )


@dataclass(frozen=True)
class SaVolume(StockAnalysis):
    volume_threshold: int = 30_000
    estimate_filter_name: str = "volume"

    def _validate(self, data: dict):
        if "volume" not in data.keys():
            raise KeyError()

    def _estimate(self, data: dict) -> float:
        return 1 if data["volume"] >= self.volume_threshold else 0



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, NoReturn, Optional, Union

import pandas as pd

from kabutobashi.domain.errors import KabutobashiEntityError
from kabutobashi.domain.services.analyze import StockAnalysis
from kabutobashi.domain.services.method import Method
from kabutobashi.domain.values import StockDataEstimated, StockDataProcessed, StockDataVisualized, StockRecordset

__all__ = ["StockCodeSingleAggregate", "IStockCodeSingleAggregateRepository"]


@dataclass(frozen=True)
class StockCodeSingleAggregate:
    """
    StockCodeSingleAggregate: Aggregate

    Examples:
        >>> import kabutobashi as kb
        >>> import pandas as pd
        >>> data_list = []
        >>> records = kb.example()
        >>> methods = kb.methods + [kb.basic, kb.pct_change, kb.volatility]
        >>> stock_analysis = kb.stock_analysis
        >>>
        >>> for df in records.to_code_iterable(until=1, row_more_than=80):
        >>>     agg = kb.StockCodeSingleAggregate.of(entity=df).with_processed(methods).with_estimated(stock_analysis)
        >>>     print(agg.weighted_estimated_value())
    """

    code: str
    single_recordset: StockRecordset
    processed_list: List[StockDataProcessed] = field(default_factory=list, repr=False)
    estimated_list: List[StockDataEstimated] = field(default_factory=list, repr=False)

    @staticmethod
    def of(entity: Union[pd.DataFrame, StockRecordset], *, code: Optional[str] = None) -> "StockCodeSingleAggregate":
        if type(entity) is pd.DataFrame:
            single_recordset = StockRecordset.of(df=entity)
        elif type(entity) is StockRecordset:
            if code is None:
                raise KabutobashiEntityError("code is required")
            df_ = entity.to_df(code=code)
            single_recordset = StockRecordset.of(df=df_)
        else:
            raise KabutobashiEntityError("accept pd.DataFrame or StockDataSingleCode")

        code = single_recordset.get_single_code_recordset_status().code
        return StockCodeSingleAggregate(code=code, single_recordset=single_recordset)

    def with_processed(self, methods: List[Method]) -> "StockCodeSingleAggregate":
        # check all methods
        for method in methods:
            if not isinstance(method, Method):
                raise KabutobashiEntityError()

        df = self.single_recordset.to_df()
        return StockCodeSingleAggregate(
            code=self.code,
            single_recordset=self.single_recordset,
            processed_list=[m.process_method.process(df=df) for m in methods],
        )

    def _to_single_estimated(self, stock_analysis: StockAnalysis) -> StockDataEstimated:
        def get_impacts(influence: int = 2, tail: int = 5) -> Dict[str, float]:
            data_ = {}
            for a in self.processed_list:
                data_.update(a.get_impact(influence=influence, tail=tail))
            return data_

        def get_parameters():
            data_ = {}
            for a in self.processed_list:
                data_.update(a.parameters)
            return data_

        data = {}
        data.update(get_impacts())
        data.update(get_parameters())
        return StockDataEstimated(
            code=self.code,
            estimate_filter_name=stock_analysis.estimate_filter_name,
            estimated_value=stock_analysis.estimate(data=data),
        )

    def with_estimated(self, stock_analysis: List[StockAnalysis]) -> "StockCodeSingleAggregate":
        """

        Returns:
            StockDataEstimatedByMultipleFilter
        """
        return StockCodeSingleAggregate(
            code=self.code,
            single_recordset=self.single_recordset,
            processed_list=self.processed_list,
            estimated_list=[self._to_single_estimated(stock_analysis=sa) for sa in stock_analysis],
        )

    def weighted_estimated_value(self, weights: dict = None) -> float:
        if not weights:
            weights = {}
        return sum([e.weighted_estimated_value(weights=weights) for e in self.estimated_list])

    def estimate_filter_concat_name(self) -> str:
        estimate_filter_names = sorted([e.estimate_filter_name for e in self.estimated_list])
        return "_".join(estimate_filter_names)

    def visualize(self, method: Method, size_ratio: int = 2) -> StockDataVisualized:
        df = self.single_recordset.to_df()
        processed_df = method.process_method.process(df=df)
        if method.visualize_method:
            return method.visualize_method.visualize(size_ratio=size_ratio, df=processed_df.df)
        else:
            raise KabutobashiEntityError(f"method {method.process_method.method_name} has no visualize method")


class IStockCodeSingleAggregateRepository(ABC):
    def read(self, code: str) -> "StockCodeSingleAggregate":
        return StockCodeSingleAggregate(
            code=code,
            single_recordset=self._stock_data_read(code=code),
            processed_list=self._stock_processed_read(code=code),
            estimated_list=self._stock_estimated_read(code=code),
        )

    @abstractmethod
    def _stock_data_read(self, code: str) -> "StockRecordset":
        raise NotImplementedError()

    @abstractmethod
    def _stock_processed_read(self, code: str) -> List["StockDataProcessed"]:
        raise NotImplementedError()

    @abstractmethod
    def _stock_estimated_read(self, code: str) -> List["StockDataEstimated"]:
        raise NotImplementedError()

    def write(self, data: StockCodeSingleAggregate) -> NoReturn:
        self._stock_data_write(data=data.single_recordset)
        self._stock_processed_write(data=data.processed_list)
        self._stock_estimated_write(data=data.estimated_list)

    @abstractmethod
    def _stock_data_write(self, data: StockRecordset) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def _stock_processed_write(self, data: List[StockDataProcessed]) -> NoReturn:
        raise NotImplementedError()

    @abstractmethod
    def _stock_estimated_write(self, data: List[StockDataEstimated]) -> NoReturn:
        raise NotImplementedError()



from .analyze import SaFundamental, SaVolume, StockAnalysis
from .decode_pages import (
    PageDecoder,
    StockInfoHtmlDecoder,
    StockInfoMultipleDaysHtmlDecoder,
    StockIpoHtmlDecoder,
    Weeks52HighLowHtmlDecoder,
)



"""
define structures of the stock-data,
when processing Methods like SMA, MCAD,
and when estimating stock-code which is to rise in the next day or so on.

- Used for ``crawling``

  - StockIpo
  - Weeks52HighLow

- define data-structure: ``basement``

  - StockRecordset
  - StockCodeSingleAggregate

- initial step to analyze:  ``processed``

  - StockDataProcessed

- second step to analyze:  ``estimated``

  - StockDataEstimated
"""
from .stock_models import OPTIONAL_COL, REQUIRED_COL, StockBrand, StockIpo, StockRecord, Weeks52HighLow



from dataclasses import dataclass

import pandas as pd

from .method import Method, MethodType, ProcessMethod, VisualizeMethod


@dataclass(frozen=True)
class BollingerBandsProcess(ProcessMethod):
    """
    株価の勢いの変化や反転の目安、方向を見る ``BollingerBands`` を計算するクラス。

    See Also:
        * https://www.sevendata.co.jp/shihyou/technical/bori.html
    """

    band_term: int = 12
    continuity_term: int = 10
    method_name: str = "bollinger_bands"
    method_type: MethodType = MethodType.TECHNICAL_ANALYSIS

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(mean=df["close"].rolling(self.band_term).mean(), std=df["close"].rolling(self.band_term).std())
        df = df.assign(
            upper_1_sigma=df.apply(lambda x: x["mean"] + x["std"] * 1, axis=1),
            lower_1_sigma=df.apply(lambda x: x["mean"] - x["std"] * 1, axis=1),
            upper_2_sigma=df.apply(lambda x: x["mean"] + x["std"] * 2, axis=1),
            lower_2_sigma=df.apply(lambda x: x["mean"] - x["std"] * 2, axis=1),
            upper_3_sigma=df.apply(lambda x: x["mean"] + x["std"] * 3, axis=1),
            lower_3_sigma=df.apply(lambda x: x["mean"] - x["std"] * 3, axis=1),
        )
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(
            over_upper=df.apply(lambda x: 1 if x["close"] > x["upper_2_sigma"] else 0, axis=1),
            over_lower=df.apply(lambda x: 1 if x["close"] < x["lower_2_sigma"] else 0, axis=1),
            over_upper_continuity=lambda x: x["over_upper"].rolling(self.continuity_term).sum(),
            over_lower_continuity=lambda x: x["over_lower"].rolling(self.continuity_term).sum(),
        )

        df["buy_signal"] = df["over_upper"].apply(lambda x: 1 if x > 0 else 0)
        df["sell_signal"] = df["over_lower"].apply(lambda x: 1 if x > 0 else 0)
        return df

    def _processed_columns(self) -> list:
        return [
            "upper_1_sigma",
            "lower_1_sigma",
            "upper_2_sigma",
            "lower_2_sigma",
            "upper_3_sigma",
            "lower_3_sigma",
            "over_upper_continuity",
            "over_lower_continuity",
        ]

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return {
            "upper_1_sigma": df_p["upper_1_sigma"].tail(3).mean(),
            "lower_1_sigma": df_p["lower_1_sigma"].tail(3).mean(),
            "upper_2_sigma": df_p["upper_2_sigma"].tail(3).mean(),
            "lower_2_sigma": df_p["lower_2_sigma"].tail(3).mean(),
        }


@dataclass(frozen=True)
class BollingerBandsVisualize(VisualizeMethod):
    """
    株価の勢いの変化や反転の目安、方向を見る ``BollingerBands`` を計算するクラス。

    See Also:
        * https://www.sevendata.co.jp/shihyou/technical/bori.html
    """

    def _color_mapping(self) -> list:
        return [
            {"df_key": "upper_1_sigma", "color": "#dc143c", "label": "+1s"},
            {"df_key": "lower_1_sigma", "color": "#dc143c", "label": "-1s"},
            {"df_key": "upper_2_sigma", "color": "#ffa500", "label": "+2s"},
            {"df_key": "lower_2_sigma", "color": "#ffa500", "label": "-2s"},
            {"df_key": "upper_3_sigma", "color": "#1e90ff", "label": "+3s"},
            {"df_key": "lower_3_sigma", "color": "#1e90ff", "label": "-3s"},
        ]

    def _visualize_option(self) -> dict:
        return {"position": "in"}


bollinger_bands = Method.of(process_method=BollingerBandsProcess(), visualize_method=BollingerBandsVisualize())



from dataclasses import dataclass

import pandas as pd

from kabutobashi.domain.errors import KabutobashiMethodError

from .method import Method, MethodType, ProcessMethod


@dataclass(frozen=True)
class IndustryCategoriesProcess(ProcessMethod):
    """
    株のvolumeやPBR, PSR, PERなどの値を返す。
    parameterizeのみに利用される。
    """

    method_name: str = "industry_categories"
    method_type: MethodType = MethodType.PARAMETERIZE
    INDUSTRY_TYPE_MAPPING = {
        "水産・農林業": "industry_fisheries_agriculture",
        "鉱業": "industry_mining",
        "建設業": "industry_construction",
        "食料品": "industry_food",
        "繊維製品": "industry_fiber",
        "パルプ・紙": "industry_pulp_paper",
        "化学": "industry_chemistry",
        "医薬品": "industry_pharmaceuticals",
        "石油・石炭製品": "industry_oil_coal",
        "ゴム製品": "industry_rubber",
        "ガラス・土石製品": "industry_glass",
        "鉄鋼": "industry_steel",
        "非鉄金属": "industry_non_ferrous_metals",
        "金属製品": "industry_metal_products",
        "機械": "industry_machine",
        "電気機器": "industry_electric",
        "輸送用機器": "industry_transportation",
        "精密機器": "industry_mechanical_equipment",
        "その他製品": "industry_other",
        "電気・ガス業": "industry_electricity_gas",
        "陸運業": "industry_land_transportation",
        "海運業": "industry_shipping",
        "空運業": "industry_air_freight",
        "倉庫・運輸関連業": "industry_warehouse",
        "情報・通信業": "industry_information",
        "卸売業": "industry_wholesale",
        "小売業": "industry_retail",
        "銀行業": "industry_back",
        "証券、商品先物取引業": "industry_stock_future",
        "保険業": "industry_insurance",
        "その他金融業": "industry_financial",
        "不動産業": "industry_real_state",
        "サービス業": "industry_service",
    }

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["diff"] = -1
        # 正負が交差した点
        df_ = df_.join(self._cross(df_["diff"]))
        df_ = df_.rename(columns={"to_plus": "buy_signal", "to_minus": "sell_signal"})
        return df_

    def _processed_columns(self) -> list:
        return []

    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        params = {v: 0 for v in self.INDUSTRY_TYPE_MAPPING.values()}
        industry_type_list = list(df_x["industry_type"].unique())
        if len(industry_type_list) > 1:
            raise KabutobashiMethodError("industry type should be 1")
        key = industry_type_list[0]
        if key in self.INDUSTRY_TYPE_MAPPING.keys():
            params.update({self.INDUSTRY_TYPE_MAPPING[key]: 1})
        return params


industry_categories = Method.of(process_method=IndustryCategoriesProcess(), visualize_method=None)



from datetime import datetime
from typing import Optional, Union

from pydantic import BaseModel, Field

from kabutobashi.domain.errors import KabutobashiEntityError

REQUIRED_COL = ["code", "open", "close", "high", "low", "volume", "per", "psr", "pbr", "dt"]
OPTIONAL_COL = ["name", "industry_type", "market", "unit", "is_delisting"]


__all__ = ["StockBrand", "StockRecord", "StockIpo", "Weeks52HighLow", "REQUIRED_COL", "OPTIONAL_COL"]


def _replace(input_value: str) -> str:
    if input_value == "-":
        return "0"
    return input_value.replace("---", "0").replace("円", "").replace("株", "").replace("倍", "").replace(",", "")


def _convert_float(input_value: Union[str, float, int]) -> float:
    if type(input_value) is float:
        return input_value
    elif type(input_value) is int:
        return float(input_value)
    elif type(input_value) is str:
        try:
            return float(_replace(input_value=input_value))
        except ValueError as e:
            raise KabutobashiEntityError(f"cannot convert {input_value} to float: {e}")
    raise KabutobashiEntityError(f"cannot convert {input_value} to float")


def _convert_int(input_value: Union[str, float, int]) -> int:
    if type(input_value) == int:
        return input_value
    elif type(input_value) == float:
        try:
            return int(input_value)
        except ValueError:
            return 0
    elif type(input_value) is str:
        try:
            return int(_replace(input_value=input_value))
        except ValueError as e:
            raise KabutobashiEntityError(f"cannot convert {input_value} to integer: {e}")
    raise KabutobashiEntityError(f"cannot convert {input_value} to int")


class StockBrand(BaseModel):
    """
    StockBrand: entity
    """

    id: int = Field(description="id")
    code: str = Field(description="銘柄コード")
    unit: int = Field(description="単位")
    market: str = Field(description="市場")
    name: str = Field(description="銘柄名")
    industry_type: str = Field(description="業種")
    market_capitalization: str = Field(description="時価総額")
    issued_shares: str = Field(description="発行済み株式")

    def __init__(
        self,
        id: Optional[int],
        code: str,
        unit: int,
        market: str,
        name: str,
        industry_type: str,
        market_capitalization: str,
        issued_shares: str,
    ):
        if id is None:
            id = 0
        # code may "100.0"
        code = code.split(".")[0]

        super().__init__(
            id=id,
            code=code,
            unit=unit,
            market=market,
            name=name,
            industry_type=industry_type,
            market_capitalization=market_capitalization,
            issued_shares=issued_shares,
        )

    def dumps(self) -> dict:
        return self.dict()

    @staticmethod
    def loads(data: dict) -> "StockBrand":
        # code may "100.0"
        code = str(data["code"]).split(".")[0]

        return StockBrand(
            id=data.get("id"),
            code=code,
            unit=_convert_int(data.get("unit", 0)),
            market=data.get("market", ""),
            name=data.get("name", ""),
            industry_type=data.get("industry_type", ""),
            market_capitalization=data.get("market_capitalization", ""),
            issued_shares=data.get("issued_shares", ""),
        )

    def is_reit(self) -> bool:
        return self.market == "東証REIT"

    def __eq__(self, other):
        if not isinstance(other, StockBrand):
            return False
        return self.code == other.code

    def __hash__(self):
        return hash(self.code)

    class Config:
        orm_mode = True


class StockRecord(BaseModel):
    """
    StockRecord: entity
    """

    id: int = Field(description="id")
    code: str = Field(description="銘柄コード")
    open: float = Field(description="始値")
    high: float = Field(description="高値")
    low: float = Field(description="底値")
    close: float = Field(description="終値")
    psr: float = Field(description="PSR")
    per: float = Field(description="PER")
    pbr: float = Field(description="PBR")
    volume: int = Field(description="出来高")
    is_delisting: bool = Field(description="上場廃止")
    dt: str = Field(description="日付")

    def __init__(
        self,
        id: Optional[int],
        code: str,
        open: float,
        high: float,
        low: float,
        close: float,
        psr: float,
        per: float,
        pbr: float,
        volume: int,
        is_delisting: bool,
        dt: str,
    ):

        if id is None:
            id = 0
        super().__init__(
            id=id,
            code=code,
            open=open,
            high=high,
            low=low,
            close=close,
            psr=psr,
            per=per,
            pbr=pbr,
            volume=volume,
            is_delisting=is_delisting,
            dt=dt,
        )

    @staticmethod
    def from_page_of(data: dict) -> "StockRecord":
        label_split = data["stock_label"].split("  ")
        try:
            return StockRecord(
                id=data.get("id"),
                code=label_split[0],
                open=_convert_float(data["open"]),
                high=_convert_float(data["high"]),
                low=_convert_float(data["low"]),
                close=_convert_float(data["close"]),
                psr=_convert_float(data["psr"]),
                per=_convert_float(data["per"]),
                pbr=_convert_float(data["pbr"]),
                volume=_convert_int(data["volume"]),
                is_delisting=data.get("is_delisting", False),
                dt=data["date"],
            )
        except Exception:
            return StockRecord(
                id=data.get("id"),
                code="",
                open=0,
                high=0,
                low=0,
                close=0,
                psr=0,
                per=0,
                pbr=0,
                volume=0,
                is_delisting=False,
                dt="",
            )

    def is_outlier(self) -> bool:
        return self.open == 0 or self.high == 0 or self.low == 0 or self.close == 0

    def dumps(self) -> dict:
        return self.dict()

    @staticmethod
    def loads(data: dict) -> "StockRecord":

        data_date = data.get("date")
        data_dt = data.get("dt")
        data_crawl_datetime = data.get("crawl_datetime")

        if data_date and data_dt and data_crawl_datetime:
            raise KabutobashiEntityError("日付のカラム[dt, date, crawl_datetime]のいずれかしか選べません")
        if data_date:
            dt = data_date
        elif data_dt:
            dt = data_dt
        elif data_crawl_datetime:
            dt = datetime.fromisoformat(data_crawl_datetime).strftime("%Y-%m-%d")
        else:
            raise KabutobashiEntityError("日付のカラム[dt, date, crawl_datetime]のいずれかが存在しません")

        # code may "100.0"
        code = str(data["code"]).split(".")[0]
        return StockRecord(
            id=data.get("id"),
            code=code,
            open=_convert_float(data["open"]),
            high=_convert_float(data["high"]),
            low=_convert_float(data["low"]),
            close=_convert_float(data["close"]),
            psr=_convert_float(data["psr"]),
            per=_convert_float(data["per"]),
            pbr=_convert_float(data["pbr"]),
            volume=_convert_int(data["volume"]),
            is_delisting=data.get("is_delisting", False),
            dt=dt,
        )

    class Config:
        orm_mode = True


class StockIpo(BaseModel):
    """
    まだ取り込んでない値など

    '想定(仮条件)': '1,920(1,900-1,920)',
    '吸収金額': '75.6億',
    '(騰落率)損益': '(+1.1%)+2,100円00001',

    """

    id: int = Field(description="id")
    code: str = Field(description="銘柄コード")
    manager: str = Field(description="主幹")
    stock_listing_at: str = Field(description="上場日")
    public_offering: float = Field(description="公募")
    evaluation: str = Field(description="評価")
    initial_price: float = Field(description="初値")

    def __init__(
        self,
        id: Optional[int],
        code: str,
        manager: str,
        stock_listing_at: str,
        public_offering: float,
        evaluation: str,
        initial_price: float,
    ):
        if id is None:
            id = 0
        super().__init__(
            id=id,
            code=code,
            manager=manager,
            stock_listing_at=stock_listing_at,
            public_offering=public_offering,
            evaluation=evaluation,
            initial_price=initial_price,
        )

    @staticmethod
    def loads(data: dict) -> "StockIpo":
        return StockIpo(
            id=data.get("id"),
            code=data["code"],
            manager=data["主幹"],
            stock_listing_at=data["上場"],
            public_offering=_convert_float(data["公募"]),
            evaluation=data["評価"],
            initial_price=_convert_float(data["初値"]),
        )

    def dumps(self) -> dict:
        return self.dict()

    class Config:
        orm_mode = True


class Weeks52HighLow(BaseModel):
    """
    Weeks52HighLow: Entity

    52週高値・底値の値を保持するクラス
    """

    id: int = Field(description="id")
    code: str = Field(description="銘柄コード")
    brand_name: str = Field(description="銘柄名")
    # close: float = Field(description="終値")
    buy_or_sell: str = Field(description="買い, 強い買い, 売り, 強い売り")
    # volatility_ratio: float = Field(description="価格変動比")
    # volatility_value: float = Field(description="価格変動値")
    dt: str = Field(description="日付")

    @staticmethod
    def from_page_of(data: dict) -> "Weeks52HighLow":

        return Weeks52HighLow(
            id=0, code=data["code"], brand_name=data["brand_name"], buy_or_sell=data["buy_or_sell"], dt=data["dt"]
        )

    def dumps(self) -> dict:
        return self.dict()

    class Config:
        orm_mode = True



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from logging import INFO, getLogger
from typing import Generator, List, NoReturn, Optional, Set, Tuple

import pandas as pd

from kabutobashi.domain.entity import OPTIONAL_COL, REQUIRED_COL, StockBrand, StockRecord
from kabutobashi.domain.errors import KabutobashiEntityError

logger = getLogger(__name__)
logger.setLevel(INFO)


@dataclass(frozen=True)
class StockRecordsetStatus:
    code: str
    stop_updating: bool = field(metadata={"jp": "更新停止"})
    contains_outlier: bool = field(metadata={"jp": "例外値を含む"})


@dataclass(frozen=True)
class StockRecordset:
    """
    StockRecordset: root-entity
    """

    brand_set: Set[StockBrand] = field(repr=False)
    recordset: List[StockRecord] = field(repr=False)
    code_num: int = field(metadata={"jp": "銘柄数"})
    recordset_length: int = field(metadata={"jp": "データ数"})

    def __post_init__(self):
        if not self.recordset:
            raise KabutobashiEntityError(f"required stock_data")

    @staticmethod
    def of(df: pd.DataFrame) -> "StockRecordset":
        if "code" not in df.columns:
            raise KabutobashiEntityError
        recordset = []
        brand_set = set()
        df = df.dropna(subset=["code"])
        for _, row in df.iterrows():
            try:
                recordset.append(StockRecord.loads(dict(row)))
                brand_set.add(StockBrand.loads(data=dict(row)))
            except KabutobashiEntityError:
                logger.warning(f"error occurred at: {row}")
        return StockRecordset(
            brand_set=brand_set, recordset=recordset, code_num=len(brand_set), recordset_length=len(recordset)
        )

    def get_single_code_recordset_status(self) -> StockRecordsetStatus:
        def _check_recent_update(_df: pd.DataFrame) -> bool:
            """
            直近の更新が止まっているかどうか
            """
            return (
                (len(_df["open"].tail(10).unique()) == 1)
                or (len(_df["high"].tail(10).unique()) == 1)
                or (len(_df["low"].tail(10).unique()) == 1)
                or (len(_df["close"].tail(10).unique()) == 1)
            )

        if self.code_num > 1:
            raise KabutobashiEntityError

        code = list(self.brand_set)[0].code
        df = self._to_df(code=code)

        contains_outlier = any([v.is_outlier() for v in self.recordset])
        return StockRecordsetStatus(
            code=code, stop_updating=_check_recent_update(_df=df), contains_outlier=contains_outlier
        )

    def sliding_split(
        self, *, buy_sell_term_days: int = 5, sliding_window: int = 60, step: int = 3
    ) -> Generator[Tuple[int, pd.DataFrame, pd.DataFrame], None, None]:
        """
        単一の銘柄に関してwindow幅を ``sliding_window`` 日として、
        保持しているデータの期間の間をslidingしていく関数。

        Args:
            buy_sell_term_days: この日数後までデータを切り出す。
            sliding_window: slidingしていくwindow幅
            step: windowsをずらしていく期間

        Returns:
            idx: 切り出された番号。
            df_for_x: 特徴量を計算するためのDataFrame。
            df_for_y: ``buy_sell_term_days`` 後のDataFrameを返す。値動きを追うため。
        """
        if self.code_num > 1:
            raise KabutobashiEntityError
        df = self.to_df()
        df_length = len(df.index)
        if df_length < buy_sell_term_days + sliding_window:
            raise KabutobashiEntityError("入力されたDataFrameの長さがwindow幅よりも小さいです")
        loop = df_length - (buy_sell_term_days + sliding_window)
        for idx, i in enumerate(range(0, loop, step)):
            offset = i + sliding_window
            end = offset + buy_sell_term_days
            yield idx, df[i:offset], df[offset:end]

    def get_code_list(self) -> List[str]:
        return list([v.code for v in self.brand_set])

    def _to_df(self, code: Optional[str]) -> pd.DataFrame:
        df_brand = pd.DataFrame([v.dumps() for v in self.brand_set])
        if code:
            df_brand = df_brand[df_brand["code"] == code]
        df_record = pd.DataFrame([v.dumps() for v in self.recordset])
        df = pd.merge(left=df_brand, right=df_record, how="inner", on="code")

        df = df.convert_dtypes()
        # order by dt
        idx = pd.to_datetime(df["dt"]).sort_index()
        df.index = idx
        df = df.sort_index()
        return df

    def to_df(self, *, minimum=True, latest=False, code: Optional[str] = None):
        df = self._to_df(code=code)

        if latest:
            latest_dt = max(df["dt"])
            df = df[df["dt"] == latest_dt]

        if minimum:
            return df[REQUIRED_COL]
        else:
            return df[REQUIRED_COL + OPTIONAL_COL]

    def to_single_code(self, code: str) -> "StockRecordset":
        if type(code) is not str:
            raise KabutobashiEntityError(f"code must be type of `str`")
        return StockRecordset.of(df=self._to_df(code=code))

    def to_code_iterable(
        self,
        until: Optional[int] = None,
        *,
        skip_reit: bool = True,
        row_more_than: Optional[int] = 80,
        code_list: list = None,
    ) -> Generator[pd.DataFrame, None, None]:
        _count = 0
        df = self._to_df(code=None)

        if code_list:
            df = df[df["code"].isin(code_list)]
        if skip_reit:
            df = df[~(df["market"] == "東証REIT")]

        for code, df_ in df.groupby("code"):
            if row_more_than:
                if len(df_.index) < row_more_than:
                    continue

            # add counter if yield
            if until:
                if _count >= until:
                    return
            _count += 1

            yield df_


class IStockRecordsetRepository(ABC):
    def read(self) -> "StockRecordset":
        return self._stock_recordset_read()

    @abstractmethod
    def _stock_recordset_read(self) -> "StockRecordset":
        raise NotImplementedError()

    def write(self, data: StockRecordset) -> NoReturn:
        self._stock_recordset_write(data=data)

    @abstractmethod
    def _stock_recordset_write(self, data: StockRecordset) -> NoReturn:
        raise NotImplementedError()



import random
from dataclasses import dataclass


class UserAgent:
    @staticmethod
    def get_user_agent() -> str:
        user_agent_list = [Chrome().user_agent, Firefox().user_agent]
        return random.choice(user_agent_list)

    @staticmethod
    def get_user_agent_header() -> dict:
        return {"User-Agent": UserAgent.get_user_agent()}


@dataclass
class Browser:
    """
    user_agentを生成するクラス
    user_agent_formatとbrowser_versionを受け取り、USER_AGENTを生成する
    """

    user_agent_format: str
    browser_version: str

    def __post_init__(self):
        self.user_agent = self.user_agent_format.format(version=self.browser_version)


class Chrome:
    """
    Google ChromeのUSER_AGENTを生成するクラス
    """

    def __init__(self):
        os_format_list = [
            # mac_chrome_format
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{version} Safari/537.36",
            # windows_chrome_format
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{version}",
        ]

        version_list = ["77.0.3865", "78.0.3904", "79.0.3945.130", "80.0.3987", "81.0.4044", "83.0.4103.97"]

        browser = Browser(random.choice(os_format_list), random.choice(version_list))
        self.user_agent = browser.user_agent


class Firefox:
    """
    FirefoxのUSER_AGENTを生成するクラス
    """

    def __init__(self):
        os_format_list = [
            # mac_firefox_format
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:62.0) Gecko/20100101 Firefox/{version}",
            # windows_firefox_format
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:61.0) Gecko/20100101 Firefox/{version}",
        ]

        version_list = ["68.4.2", "72.0.2", "73.0", "74.0"]

        browser = Browser(random.choice(os_format_list), random.choice(version_list))
        self.user_agent = browser.user_agent
